/*
** owner v1.2.4
**
** Run command as file owner
**
** Copyright (C) 3.12.1996 by Andreas Ley <andreas.ley@kit.edu>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** This program has been tested on a HP9000/715 with HP-UX A.09.07
** In this environment, neither lint -u nor gcc -Wall -fsyntax-only produce
** any messages. If you encounter any errors or need to make any changes to
** port it to another platform, please contact me.
**
** Version history
**	If both -o and -e specify the same file, redirect both STDOUT and STDERR into the same file
**
** Version 1.2.4 - 2.7.2020
**	???
**
** Version 1.2.3 - 13.1.2014
**	Added -O and -E options
**
** Version 1.2.2 - 24.4.2013
**	Added -l option
**
** Version 1.2.1 - 14.2.2013
**	Added -c option
**
** Version 1.2 - 28.1.2013
**	Added -o and -e options
**
** Version 1.1 - 8.3.2011
**	Added -u option
**
** Version 1.0.1 - 19.12.2007
**	Added -i option
**
** Version 1.0 - 3.12.1996
**	Initial version
*/

static char copyright[] = "@(#)Copyright (C) 1996,2020-2014 by Andreas Ley (Andreas.Ley@kit.edu)";
static char sccsid[] = "@(#)owner v1.2.4 - Run command as file owner";


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>

#ifndef FALSE
#define	FALSE	(0)		/* This is the naked Truth */
#define	TRUE	(1)		/* and this is the Light */
#endif

char		*image;
int		debug=0,ignore_errors=FALSE,set_home=FALSE,explicit_user=FALSE;


static void usage()
{
	(void)fprintf(stderr,"Usage: %s [-i] [-l] [-c directory] [{-o|-O} filename] [{-e|-E} filename] <filename> <command>\n",image);
	(void)fprintf(stderr,"or:    %s [-i] [-l] [-c directory] [-o filename] [-e filename] -u uid:gid <command>\n",image);
	(void)fprintf(stderr,"-i  Ignore non-fatal errors (e.g. setting additional groups)\n");
	(void)fprintf(stderr,"-l  Set HOME to login directory\n");
	(void)fprintf(stderr,"-c  Change working directory after changing identity but before executing command\n");
	(void)fprintf(stderr,"-o  Redirect stdout to file after changing identity but before executing command\n");
	(void)fprintf(stderr,"-O  Redirect stdout to append to file after changing identity but before executing command\n");
	(void)fprintf(stderr,"-e  Redirect stderr to file after changing identity but before executing command\n");
	(void)fprintf(stderr,"-E  Redirect stderr to append to file after changing identity but before executing command\n");
	exit(1);
}


int main(argc,argv)
int	argc;
char	*argv[];
{
	int		c,fd,flags_stdout,flags_stderr;
	extern char	*optarg;
	extern int	optind;
	char		*ptr,*cwd=NULL,*redirect_stdout=NULL,*redirect_stderr=NULL;
	struct stat	buf;

	if ((image=strrchr(argv[0],'/')))
		image++;
	else
		image=argv[0];

	while ((c=getopt(argc,argv,"Dvhilc:o:O:e:E:u:?"))!=EOF)
		switch ((char)c) {
		case 'D':
			debug++;
			break;
		case 'i':
			ignore_errors=TRUE;
			break;
		case 'l':
			set_home=TRUE;
			break;
		case 'c':
			cwd=optarg;
			break;
		case 'o':
			if (redirect_stdout)
				usage();
			redirect_stdout=optarg;
			flags_stdout=O_WRONLY|O_CREAT|O_TRUNC;
			break;
		case 'O':
			if (redirect_stdout)
				usage();
			redirect_stdout=optarg;
			flags_stdout=O_WRONLY|O_CREAT|O_APPEND;
			break;
		case 'e':
			if (redirect_stderr)
				usage();
			redirect_stderr=optarg;
			flags_stderr=O_WRONLY|O_CREAT|O_TRUNC;
			break;
		case 'E':
			if (redirect_stderr)
				usage();
			redirect_stderr=optarg;
			flags_stderr=O_WRONLY|O_CREAT|O_APPEND;
			break;
		case 'u':
			explicit_user=TRUE;
			/* FIXME: should handle user names as well as numeric ids */
			if ((ptr=strpbrk(optarg,":."))) {
				*ptr++='\0';
				buf.st_gid=atoi(ptr);
			}
			else
				usage();
			buf.st_uid=atoi(optarg);
			break;
		case 'v':
			(void)fprintf(stderr,"%s\n",sccsid+4);
			(void)fprintf(stderr,"%s\n",copyright+4);
			exit(0);
		case 'h':
			(void)fprintf(stderr,"%s\n",sccsid+4);
			(void)fprintf(stderr,"%s\n",copyright+4);
		case '?':
			usage();
		}

	if (optind+(explicit_user?1:2)>argc)
		usage();

	if (!explicit_user) {
		if (stat(argv[optind],&buf)) {
			(void)fprintf(stderr,"%s: can't stat ",image);
			perror(argv[optind]);
			exit(1);
		}
		optind++;
	}

	if (setgroups(0,NULL)<0 && !ignore_errors) {
		(void)fprintf(stderr,"%s: can't set groups: ",image);
		perror("");
		exit(1);
	}

	if (setgid(buf.st_gid)<0) {
		(void)fprintf(stderr,"%s: can't set gid: ",image);
		perror("");
		exit(1);
	}

	if (setuid(buf.st_uid)<0) {
		(void)fprintf(stderr,"%s: can't set uid: ",image);
		perror("");
		exit(1);
	}

	if (cwd) {
		if (chdir(cwd)) {
			(void)fprintf(stderr,"%s: can't chdir: ",image);
			perror(cwd);
			exit(1);
		}
	}

	if (redirect_stdout) {
		if ((fd=open(redirect_stdout,flags_stdout,0666))<0) {
			(void)fprintf(stderr,"%s: can't open: ",image);
			perror(redirect_stdout);
			exit(1);
		}
		dup2(fd,STDOUT_FILENO);
	}

	if (redirect_stderr) {
		/* If both -o and -e specify the same file, reuse fd to redirect both STDOUT and STDERR into the same file */
		if (!redirect_stdout || strcmp(redirect_stdout,redirect_stderr)) {
			if ((fd=open(redirect_stderr,flags_stderr,0666))<0) {
				(void)fprintf(stderr,"%s: can't open: ",image);
				perror(redirect_stderr);
				exit(1);
			}
		}
		dup2(fd,STDERR_FILENO);
	}

	if (set_home) {
		struct passwd	*pwd;
		char		*home;
		if (!(pwd=getpwuid(buf.st_uid))) {
			(void)fprintf(stderr,"%s: can't getpwuid(%d",image,buf.st_uid);
			perror(")");
			exit(1);
		}
		if (debug)
			(void)fprintf(stderr,"%s: getpwuid(%d)->pw_dir=\"%s\"\n",image,buf.st_uid,pwd->pw_dir);
		if (!(home=malloc(strlen(pwd->pw_dir)+6))) {
			(void)fprintf(stderr,"%s: can't malloc(%ld)",image,strlen(pwd->pw_dir)+6);
			perror(")");
			exit(1);
		}
		sprintf(home,"HOME=%s",pwd->pw_dir);
		putenv(home);
	}

	if (debug)
		for (c=optind;c<argc;c++)
			(void)fprintf(stderr,"%s: argv[%d]=\"%s\"\n",image,c-optind,argv[c]);
	execvp(argv[optind],&argv[optind]);

	(void)fprintf(stderr,"%s: can't exec ",image);
	perror(argv[optind]);
	exit(1);
}
